﻿:Class MildServer

    :Field Public Port←8080
    :Field Public Root←'c:\MildServer\'
    :Field Public Address←'http://localhost'
    :Field Public DefaultPage←'index.htm'
    :Field Public DefaultExtension←'.dyalog'

    :Field Public RootCertDir←''
    :Field Public CertFile←''
    :Field Public KeyFile←''
    :Field Public Secure←0

    :Field Public SSLFlags←32+64  ⍝ Accept Without Validating, RequestClientCertificate
    :Field Public IPVersion←'IPv4'      ⍝ Default until further notice
    :Field Public TID←¯1                ⍝ Thread ID Server is running under
    :Field Public BlockSize←640000      ⍝ Blocksize when returning files

    :Field Public SessionHandler←⎕NS ''
    :Field Public Authentication←⎕NS ''

    NL←(CR LF)←⎕UCS 13 10
    BOM←⎕UCS 65279
    FindFirst←{(⍺⍷⍵)⍳1}                       
 
    ∇ r←Srv x
      :If 3=⎕NC'#.DRC.Srv' ⋄ r←#.DRC.Srv x,⊂'Protocol'IPVersion ⍝ Conga v2.0+
      :Else ⋄ r←#.DRC.Server(2≠⍳⍴x)/x ⋄ :EndIf                ⍝ V1.0: Remove address
    ∇
   
    ∇ r←Clt x
      :If 3=⎕NC'#.DRC.Clt' ⋄ r←#.DRC.Clt x,⊂'Protocol'IPVersion
      :Else ⋄ r←#.DRC.Client x ⋄ :EndIf
    ∇
   
    ∇ Run
      :Access Public
      ('Already Running on Thread',⍕TID)⎕SIGNAL(TID∊⎕TNUMS)/11
      TID←RunServer&⍬
    ∇

    ∇ End
      :Access Public
      :If 0=⎕NC'ServerName'
          {}1 #.DRC.Init''
      :Else
          'Not running'⎕SIGNAL(#.DRC.Exists ServerName)↓11
          {}#.DRC.Close ServerName
          {}⎕TSYNC TID
      :EndIf
      TID←¯1
    ∇

    ∇ HandleMSP REQ;inst;class;z;file;props;lcp;args;i;ts;date;n;expired;data;m;oldinst;names;html
    ⍝ Handle a "Mildserver Page" request
 :Trap 999 ⍝ Trap all errors
      :If ~#.Files.Exists file←Root,1↓REQ.Page
          REQ.Fail 404 ⍝ Not found
      :Else
          date←4⊃,#.Files.DirX file
          :If 0≠⍴REQ.Session.Pages     ⍝ Look for existing Page in Session
          :AndIf (n←⍴REQ.Session.Pages)≥i←REQ.Session.Pages._PageName⍳⊂REQ.Page
              inst←i⊃REQ.Session.Pages ⍝ Get existing instance
              :If expired←inst._PageDate≢date  ⍝ Timestamp unchanged?
                  oldinst←inst
                  REQ.Session.Pages~←inst
                  ⎕←'Page: ',REQ.Page,' ... has been updated ...'
              :EndIf
          :AndIf ~expired
     
          :Else                        ⍝ First use of Page in this Session, or page expired
              :If 0≠⍴z←#.Files.GetText file
                  class←⎕SE.SALT.Load file,' -target=#'
                  :Trap 11
                      inst←⎕NEW class
                  :Else
                      REQ.Fail 500 ⋄ →0
                  :EndTrap
                  inst._PageName←REQ.Page
                  inst._PageDate←date
                  REQ.Session.Pages,←inst
                  :If 0≠⎕NC'oldinst'
                  :AndIf (names←oldinst.⎕NL-2)≡inst.⎕NL-2 ⍝ Interface is indentical
                  :AndIf 0≠⍴names←('_'≠1⊃¨names)/names
                      ⎕←'    (interface is identical: data transferred)'
                      ⍎'inst.(',names,')←oldinst.(',(names←⍕names),')' ⍝ Transfer values
                  :EndIf
              :Else
                  REQ.Fail 404 ⋄ →0
              :EndIf
          :EndIf
     
     ⍝ Move arguments / parameters into Public Properties
         :If 0≠1↑⍴data←{⍵[⍋↑⍵[;1];]}REQ.Arguments⍪REQ.Data
             m←1,2≢/data[;1]
             data←(m/data[;1]),[1.5]m⊂data[;2]
             i←{⍵/⍳⍴⍵}1=⊃∘⍴¨data[;2]
             data[i;2]←⊃¨data[i;2]
         :EndIf
         :If 0≠⍴lcp←props←('_'≠1⊃¨props)/props←(inst.⎕NL-2) ⍝ Get list of public properties (not lowercase)
         :AndIf 0≠1↑⍴args←(data[;1]∊lcp)⌿data
             args←(2⌈⍴args)⍴args
             i←lcp⍳args[;1]
             ⍎'inst.(',(⍕props[i]),')←args[;2]'
         :EndIf
         inst.Render REQ

     :EndIf
 :Else                ⍝ Trapped error
     Error REQ
 :EndTrap
∇

    ∇ r←RunServer arg;Common;cmd;name;port;wres;ref;nspc;sink;Stop;secure;certpath;flags;z;rc
      ⍝ Simple HTTP (Web) Server framework
      ⍝ Assumes Conga available in #.DRC and uses #.HTTPRequest
      ⍝ arg; HOME port name
      ⍝ HOME: Root of file system OR a function which will handle the requests
      ⍝ certs: RootCertDir, ServerCert, ServerKey (optional: runs Secure server)
     
      :If Secure
          {}#.DRC.SetProp'.' 'RootCertDir'(1⊃certs)
          →(0≠1⊃r←Srv'' ''Port'Text' 10000 CertFile KeyFile SSLFlags)⍴0
          ⎕←'Starting secure server using certificate ',2⊃certs
      :Else
          →(0≠1⊃r←Srv'' ''Port'Text' 10000)⍴0
      :EndIf
     
      ServerName←2⊃r
      ⎕←'Web server ''',ServerName,''' started on port ',⍕Port
      ⎕←'Root folder: ',Root
     
      Common←⎕NS'' ⋄ Stop←0
     
      :While ~Stop
          wres←#.DRC.Wait ServerName 1000 ⍝ Tick every secs
          ⍝ wres: (return code) (object name) (command) (data)
     
          :Select 1⊃wres
          :Case 0 ⍝ Good data from RPC.Wait
              :Select 3⊃wres
     
              :Case 'Error'
                  :If ServerName≡2⊃wres
                      Stop←1
                  :EndIf
                  ⎕←'Error ',(⍕4⊃wres),' on ',2⊃wres
                  ⎕EX SpaceName 2⊃wres
     
              :CaseList 'Block' 'BlockLast'
     
                  :If 0=⎕NC nspc←SpaceName 2⊃wres
                      nspc ⎕NS''
                      :If Secure
                          rc z←#.DRC.GetProp(2⊃wres)'PeerCert'
                          :If rc=0 ⋄ (⍎nspc).PeerCert←z
                          :Else ⋄ ∘ ⋄ :EndIf
                      :EndIf
     
                  :EndIf
                  r←(⍎nspc)HandleRequest&wres[2 4] ⍝ Run page handler in new thread
     
                  :If 'BlockLast'≡3⊃wres
                      ⎕EX nspc
                  :EndIf
     
              :Case 'Connect' ⍝ Ignore
     
              :Else
                  ⎕←'Error ',⍕wres
              :EndSelect
     
          :Case 100 ⍝ Time out - put "housekeeping" code here
              SessionHandler.HouseKeeping ⎕THIS
     
          :Case 1010 ⍝ Object Not found
              ⎕←'Object ''',ServerName,''' has been closed - Web Server shutting down'
              →0
     
          :Else
              ⎕←'#.RPC.Wait failed:'
              ⎕←wres
              ∘
     
          :EndSelect
     
      :EndWhile
      {}#.DRC.Close name
      ⎕←r←'Web server ''',name,''' stopped '
    ∇

    ∇ Make(port root address)
      :Access Public
      :Implements Constructor
     
      SessionHandler.GetSession←{}   ⍝ So we can always
      SessionHandler.HouseKeeping←{} ⍝    call these fns
      Authentication.Authenticate←{} ⍝    without worrying
     
      {}#.DRC.Init''
      Port←port
      Root←root,((¯1↑root)∊'/\')↓'\'
    ∇

    ∇ UnMake
      :Implements Destructor
      :Trap 0 ⋄ End ⋄ :EndTrap
    ∇

    ∇ conns HandleRequest arg;buf;m;Answer;obj;CMD;pos;req;Data;z;r;hdr;I;REQ;status;file;tn;length;done;offset;res;closed;sess
 ⍝ Handle a Web Server Request
      obj buf←arg
      :If 80≠⎕DR buf ⋄ buf←⎕UCS buf ⋄ :EndIf ⍝ Make it text
     
      :If 0=conns.⎕NC'Buffer'
          conns.Buffer←''
      :EndIf
      conns.Buffer,←buf
     
      pos←3+(NL,NL)FindFirst conns.Buffer
     
      :If pos>⍴conns.Buffer ⍝ Have we got everything ?
          :Return
      :ElseIf pos>I←(z←LF,'content-length:')FindFirst hdr←#.Strings.lc pos↑conns.Buffer
      :AndIf (⍴conns.Buffer)<pos+↑2⊃⎕VFI(¯1+z⍳CR)↑z←(¯1+I+⍴z)↓hdr
          :Return ⍝ a content-length was specified but we haven't yet gotten what it says to ==> go back for more
      :EndIf
     
      req←pos↑conns.Buffer
      conns.Buffer←pos↓conns.Buffer
      REQ←⎕NEW #.HTTPRequest(req conns.Buffer)
      REQ.Server←⎕THIS ⍝ Request will also contain reference to the Server
     
      :If 2=conns.⎕NC'PeerCert' ⋄ REQ.PeerCert←conns.PeerCert ⋄ :EndIf       ⍝ Add Client Cert Information
      SessionHandler.GetSession REQ
      Authentication.Authenticate REQ
     
      REQ.Page←DefaultPage{∧/⍵∊'/\':'/',⍺ ⋄ ⍵}REQ.Page
      REQ.Page,←(~'.'∊REQ.Page)/DefaultExtension
      :If REQ.Response.Status≠401 ⍝ Authentication did not fail
          :If (¯7↑REQ.Page)≢'.dyalog'
              :If REQ.Command≡'get'
                  REQ.ReturnFile Root,REQ.Page
              :Else
                  REQ.Fail 503 ⍝ Service Unavailable
              :EndIf
     
          :Else ⍝ "MildServer Page"
              HandleMSP REQ
          :EndIf
      :EndIf
     
      :If ~REQ.Response.File ⋄ Wrap REQ ⋄ :EndIf
      res←REQ.Response
     
      :If 1=res.File ⍝ See HTTPRequest.ReturnFile
          :If #.Files.Exists file←res.HTML
              tn←file ⎕NTIE 0
              res.HTML←⎕NREAD tn 80 BlockSize
              length←⎕NSIZE tn
          :Else
              length←res.File←0 ⋄ REQ.Fail 404 ⍝ File not found
          :EndIf
      :Else
          length←⍴res.HTML
      :EndIf
     
      status←res.((⍕Status),' ',StatusText)
      hdr←{⎕ML←3 ⋄ ∊⍵}{⍺,': ',⍵,NL}/res.Headers
      Answer←'HTTP/1.0 ',(⍕res.Status),' ',res.StatusText,NL,'Content-Length: ',(⍕length),NL,hdr,NL,res.HTML
      done←length≤offset←⍴res.HTML
     
      :Repeat
          :If ~0=1⊃z←#.DRC.Send obj Answer done ⍝ Send response and
              ⎕←'Closed socket ',obj,' due to error: ',⍕z
              done←1
          :EndIf
          closed←done
          :If ~done
              done←BlockSize>⍴Answer←⎕NREAD tn 80,BlockSize,offset
              offset+←⍴Answer
          :EndIf
      :Until closed ⍝ Everything sent
     
      :If res.File ⋄ ⎕NUNTIE tn ⋄ :EndIf
    ∇

    ∇ r←SpaceName cmd
     ⍝ Generate namespace name from command name
     
      r←'Common.C',Subst(2⊃{1↓¨('.'=⍵)⊂⍵}'.',cmd)'-=' '_∆'
    ∇

    ∇ r←Subst arg;i;m;str;c;rep
      ⍝ Substitute character c in str with rep
      str c rep←arg
     
      i←c⍳str
      m←i≤⍴c
      (m/str)←rep[m/i]
      r←str
    ∇

⍝ --- Methods which are usually overridden ---

    ∇ Wrap req
      :Access Public Overridable
    ⍝ Add formatting to HTML if necessary
    ∇

    ∇ onSessionStart req
      :Access Public Overridable
    ⍝ Process a new session
    ∇

    ∇ onSessionEnd session
      :Access Public Overridable
    ⍝ Handle the end of a session
    ∇

    ∇ Error req
      :Access Public Overridable
    ⍝ Handle trapped errors
      req.HTML←'<font face="APL385 Unicode" color="red">',(⊃,/⎕DM¨⊂'<br>'),'</font>'
      req.Fail 500 ⍝ Internal Server Error
    ∇

⍝ --- End of Overridable methods

:EndClass