﻿:Namespace Base64

    CharSet←'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/='

    ∇ rc←Encode txt;⎕IO;⎕ML;p
  ⍝p Encode messages into Base64 (MIME)
  ⍝o CHK 31.05.1999 <chk@hoesle-kienzlen.de>
  ⍝d How does base64 work? The purpose of the base64 algorithm is to
  ⍝d convert all 256 characters of the character set into the 64 characters
  ⍝d shown below, as those characters are common on all possible machines
  ⍝d through wich a mail message can pass on it's way through the Internet.
  ⍝d To do this, every 24 bits (3 8-bit characters) are split into 4 6-bit
  ⍝d chunks (=24 bits again). Those 6-bit chunks are used as an index into
  ⍝d the base64 alfabet. If the message does not contain a number of
  ⍝d characters divisible by 3, "=" characters are appended as padding after
  ⍝d encoding, to help the decoding function do it's work.
     
      ⎕IO←0 ⋄ ⎕ML←3
      ⎕SIGNAL(80≠⎕DR txt)/92
      p←(3×⌈(⍴txt)÷3)-⍴txt            ⍝ number of padding characters
      txt←,⍉(8⍴2)⊤⎕UCS txt            ⍝ convert to 8-bit binary
      txt←2⊥⍉((⌈(⍴txt)÷6),6)⍴txt,5⍴0  ⍝ re-convert every 6 bits to decimal
      rc←(CharSet,'=')[txt,p⍴64]           ⍝ index into BASE64 alfabet (+ padding)
    ∇

    ∇ rc←Decode txt;⎕IO;⎕ML;p
  ⍝p Decode messages which are encoded in Base64 (MIME)
  ⍝o CHK 31.05.1999 <chk@hoesle-kienzlen.de>
  ⍝d NOTE: for an explanation of base64 see function ∆b64enc
     
      ⎕IO←0 ⋄ ⎕ML←3
  ⍝ the base 64 alfabet (= is the padding character):
     
      txt←(txt∊CharSet)/txt       ⍝ process only BASE64 characters
      p←-6×+/txt='='         ⍝ the number of padding bits
  ⍝ decode every 6 bits of the BASE64 text into ⎕AV:
      rc←p↓,⍉(6⍴2)⊤CharSet⍳txt    ⍝ convert index to 6-bit binary
      rc←⍉((⌊(⍴rc)÷8),8)⍴rc  ⍝ make chunks of 8 bits
      rc←⎕UCS 2⊥rc           ⍝ re-convert to index into ⎕AV
    ∇

:EndNamespace